# frozen_string_literal: true

# app/hermes/models/concerns/status.rb
# app/controllers/concerns/test_concern.rb

module Hermes
  module Concerns
    # Handles statuses relationships
    module Status
      extend ActiveSupport::Concern

      included do
        self.class_variable_set(:@@status_names, [])
      end
      
      def check_status_changed
        class_name = self.class.name.demodulize.underscore
        need_to_save = false

        status_names.each do |s|
          label = s[:label]
          type = s[:type]

          if send("#{label}_changed?")
            case type
            when :single
              prev_status = Hermes::Status.where("object_id": id, status_label: label).order(created_at: :desc).limit(1).first
              if !prev_status || prev_status.title != send("#{label}")
                new_status = Hermes::Status.create(
                  object_id: id,
                  status_label: label,
                  title: send("#{label}"),
                  closing: prev_status ? prev_status.id : nil
                )
                
                set(
                  "#{label}_id": new_status.id.to_s,
                  "#{label}_updated_at": new_status.created_at,
                  "old_#{label}": prev_status ? prev_status.title : nil,
                  updated_at: DateTime.now
                )
              end
            when :multiple
              statuses = send("#{label}")
              statuses.each do |status|
                performer_id = status['performer_id']
                title = status['title']
                prev_status = Hermes::Status.where("#{class_name}_id": id, status_label: label, performer_id: performer_id).order(created_at: :desc).limit(1).first
  
                if !prev_status || prev_status.title != title
                  new_status = Hermes::Status.create(
                    object_id: id,
                    status_label: label,
                    title: title,
                    performer_id: performer_id,
                    closing: prev_status ? prev_status.id : nil
                  )
  
                  old_statuses = new_statuses = send("#{label}")
                  new_statuses.reject { |s| s['id'] == prev_status.id.to_s} if prev_status
                  new_statuses << {
                    id: new_status.id.to_s,
                    performer_id: new_status.performer_id,
                    title: new_status.title,
                    created_at: new_status.created_at,
                    closing: prev_status ? prev_status.id : nil,
                  }

                  set(
                    "#{label}_id": new_statuses,
                    "#{label}_updated_at": new_status.created_at,
                    "old_#{label}": statuses,
                    updated_at: DateTime.now
                  )
                end
              end

            end
          end
        end
      end

      def status_names
        self.class.class_variable_get(:@@status_names)
      end
      
      module ClassMethods
        def add_status(type = :single, label_name = "status", default = nil)
          self.class_variable_set(:@@status_names, self.class_variable_get(:@@status_names).push({label: label_name, type: type}))

          case type
          when :single
            field "#{label_name}", type: String, default: default
            field "#{label_name}_id", type: BSON::ObjectId
            field "#{label_name}_updated_at", type: DateTime
            field "old_#{label_name}", type: String
            index({ "#{label_name}" => 1 }, background: true)
          when :multiple
            field "#{label_name}", type: Array, default: []
            field "#{label_name}_updated_at", type: DateTime
            field "old_#{label_name}", type: Array, default: []
            index({ "#{label_name}" => 1 }, background: true)
          else
            raise ArgumentError, "To add a status the type should be :single or :multiple. #{type} isn't a valid type"
          end
        end
      end
    end
  end
end
